/*

|--------------------------------------------------------------------------|
| MLPTRV.C: Multilayer Perceptron Neural Net Training/Validation program   | 
|          (Training algorithm: Back-Propagation)                          |
|                                                                          |
|            By Cristina & Fernando de Castro - September 2001             |
|	                                                                       |
|	V1.2 (September 2001) for DJGPP compiler                               |
|  	V1.2.d (February 2013) adapted for DevC++ compiler with conio2 library |
|--------------------------------------------------------------------------|

		|---------------------------------------|
		|                                       |
		|   This program should be compiled     |
		|    with DevC++ compiler V 5.3.0.4     | 
		|      using conio2 library             |
		|                                       |
        |---------------------------------------|
*/

/***************************************************************************
*
* HEADERS:
*
****************************************************************************/
#include<stdio.h>
#include<stdlib.h>
#include<process.h>
#include<math.h>
#include<float.h>
#include<conio2.h>
#include<ctype.h>
#include<string.h>
#include<time.h>
#include<dos.h>
#include<fcntl.h>
#include<share.h>


/*****************************************************************************
*
* PROGRAM DEFINITIONS:
*
*****************************************************************************/
#define BUFSIZE 0xFFFF /* file I/O buffer size: 0xFFFF = 64Kb */
#define SMALLBUFSIZE  0x03FF /* file I/O small buffer size 0x03FF = 1Kb  */
#define MAXLINESIZE 65536 /* max line size for  GetSheetNumLines() and NumberOfFloatsInLine() */
#define MAXARGLEN 256   /* max length of an alphanumeric arg */


/*****************************************************************************
*
* COMPILING DIRECTIVES:
*
*****************************************************************************/

#pragma GCC diagnostic ignored "-Wwrite-strings"


/*****************************************************************************
*
* CONSTANTS:
*
*****************************************************************************/
#define BIAS -1 /* bias level */
#define MAXFLOAT 3.4E38 /* 32 bits DJGPP max float */
#define UMAX 88.7 /* 88.7 = ln( 3.4x10^38): 'u' variable max value to avoid
		   float overflow in function activation Phy() */
#define NALPHALINES 5 /* desired number of alphanumeric lines in parameter file */
#define NFLOATLINES 9 /* desired number of float lines in parameter file */
#define BEL 7  /* ascii BEL */
#define BLINK 128 /* conio BLINK */

/***************************************************************************
*
* DATA TYPE STRUCTURES:
*
****************************************************************************/
typedef struct Neuron {

float *W; /* synaptic weights vector (size=NumSynapses->from Top[] array) */
float *DW_1; /* previous cycle synaptic weights correction factors vector
	     (for momentum calculation purposes) (size=NumSynapses->from Top[] array) */
float d; /* neuron local gradient */
float v; /* neuron activation level (and neuron function signal y after
	    application of activation function Phy(v)) */

}NEURON;


typedef struct Layer{

NEURON *N; /* neuron array (size=NumNeurons) */
float nt;  /* layer learning rate */

}LAYER;

/*--------------------------------------------------------------------------*/

typedef struct Network {

char *ParameterFile; /* input parameter file */
unsigned NumInputNodes; /* Number of input nodes */
unsigned NumLayers; /* number of computational layers in the MLP */
unsigned *Top; /* neural network topology (architecture) array (size=NumLayers) */
float *ErrVect; /* output layer error vector */
LAYER  *L; /* layer array (size=NumLayers) */
char *TDIF; /* Training Data Input File */
char *VDIF; /* Validation Data input File */
char *FMLPSF; /* Final MLP Specification Output File */
char *MSEOF; /* Training & validation MSE Output File */
FILE *MSEOFp; /* Training & validation MSE Output File Pointer */
char *MLPO2VOF; /* MLP output_to_VDIF output file */
FILE *MLPO2VOFp; /* MLP output_to_VDIF output file */
float *LVect;   /* Layer Learning Vector */
unsigned TSSize; /* Training Set size (number of vector samples in Training Set */
unsigned VSSize; /* Validation Set size (number of vector samples in Validation Set */
char *buffer; /* file I/O speed up buffer address */
float MSE_T; /* the training MSE after the last epoch */
float MSE_V; /* the validation MSE after the last epoch */
unsigned long NumEpoch; /* the number of epochs until the last epoch */
float Phy_A; /* activation function parameter A */
float Phy_B; /* activation function parameter B */
float Alpha; /* momentum factor */
float SynIRF; /* Synapses Weight Initialization Range Factor */
float MSEChRStopCrit; /* Percent Validation MSE changing rate stopping criterion */
float MSEValStopCrit; /* Validation MSE value stopping criterion */
float **TSet; /* Training Set address */
float **VSet; /* Validation Set address */
float *Avg; /* Training Set average vector */
float *SDev; /* Training Set standard deviation vector */
double A_n,B_n; /* The desired output set D normalization parameters,
                 such that Dn=A_n*D+B_n where Dn lies in the range [-1,1] */
unsigned SmMSEChNEp2C; /* The program assumes CONVERGENCE TO A MINIMUM (hopefully the global)
if the number of nearly consecutive epochs for which the MSE is reducing in a rate  smaller
than MSEChRStopCrit is greater than  SmMSEChNEp2C */
char ValidationMode; /* if !0 the program executes the training/validation procedure, otherwise it executes only the training procedure */

}NEURALNET;

/*--------------------------------------------------------------------------*/

typedef enum {

  NO_EVENT,           /*  " ", */
  MEMORY_ERR,         /*  "Not enough memory", */
  READ_OPEN_ERR,      /*  "I can't find the file", */
  READ_ERR,           /*  "Error reading file", */
  WRITE_OPEN_ERR,     /*  "I can't open the file", */
  WRITE_ERR,          /*  "I can't write the file", */
  BUFFER_ERR,         /*  "I can't attach buffer to file", */
  SIZE_ERR,           /*  "Specified size doesn't match data size in file", */
  MAXIT_ERR,          /*  "Exceeded max number of iterations in function", */
  LOGIC_ERR,          /*  "Logic error in function", */
  READ_NNET_ERR,      /*  "Incompatible neural network in file", */
  MTXDIM_ERR,         /*  "Invalid matrix dimensions in file", */
  PARF_ERR,           /*  "Wrong parameter file", */
  GFTD_ERR,           /*  "I can't get time & date info from file", */
  MAXFLOAT_READ_ERR,  /*  "Exceeded float maximum absolute value reading file" */
  /* ---*/
  USER_INT,	      /*  "Training interrupted by user! Neural Network saved to file", */
  SMALL_MSECHG_INT,   /*  "Training terminated by small enough VMSE Change Rate with VMSE =", */
  SAT_INT,	      /*  "Training terminated due to neurons saturation at VMSE =", */
  SMALL_MSE_INT,      /*  "Training terminated by small enough VMSE =", */
  EARLY_EOF_ERR,      /*  "Found premature EOF in file ", */
  HEADER_SIZE_ERR,    /*  "Found impossible number of samples\nin header of file ", */
  ABORT_INT,          /*  "Training aborted! Neural Network not saved", */
  LNVECSIZE_ERR       /*  "Wrong layer Learning Rate Vector size in file", */

}
EVENT;

/*--------------------------------------------------------------------------*/

typedef struct ParamFileReader{

char *ParameterFile; /* parameter file */
unsigned *FloatLines; /* the array of float lines indexes */
unsigned *AlphaLines; /* the array of alpha lines indexes */
unsigned NumFloatLines; /* the number of float lines */
unsigned NumAlphaLines; /* the number of alpha lines */
char **Sheet; /* parameters file sheet matrix */
char *BufferI; /* input buffer */
FILE *IFp;     /* input filepointer */
unsigned long NumLines;  /* the number of lines read */
unsigned LineMaxLength; /* Sheet max line length */
float **FloatDat; /* float data storage sheet */
char  **AlphaDat;  /* alpha data storage sheet */
unsigned *VectSize; /* the array that contains the size of each row of FloatDat */
char AllocFlag; /* Memory allocation flag: If Allocflag!=0, memory is
                  allocated upon calling Automatic_ParamFileReader() */


}PARFR;


/*****************************************************************************
*
* MACROS:
*
*****************************************************************************/
static float sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

static double aulimarg; /* absolute maximum upper limit */
#define AULIM(a) ( fabs(aulimarg=(a)) >= MAXFLOAT ? ( aulimarg< 0.0 ? (-MAXFLOAT) : (MAXFLOAT)) : aulimarg )


/***************************************************************************
*
* FUNCTION PROTOTYPES:
*
****************************************************************************/
void InitMLP(NEURALNET *Net);
void AllocLayer(NEURALNET *Net);
void AllocNeuron(NEURALNET *Net);
float WeightGen(NEURALNET *Net,unsigned FanIn);
void InitWeight(NEURALNET *Net);
void RowShuffle(float **Mat, unsigned row);
void Help(void);
void ShowParameters(NEURALNET *Net);
float ForwardPass(NEURALNET *Net, float *InputData);
void BackwardPass(NEURALNET *Net, float *InputData);
inline float Phy(NEURALNET *Net,float *v);
inline float DPhy(NEURALNET *Net,float y);
void Quit(EVENT Evt, char *Message);
char ChangeParameterMenu(NEURALNET *Net);
void Randomize(void);
float **FloatMatCAlloc(unsigned NRows,unsigned NCols);
void FloatMatFree(float **M);
long GetFileNumSamples(FILE *Fp);
unsigned long GetFileNumLines(FILE *Fp);
float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag);
void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols);
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines);
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type);
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength);
char **CharMatCAlloc(unsigned long NRows,unsigned NCols);
unsigned NumberOfFloatsInLine(char *Line);
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector);
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag);
void NetRead(NEURALNET *Net,char *FName);
void NetWrite(NEURALNET *Net, char *FName);
void MLPTrainer(NEURALNET *Net);
void TSetDeAvg (NEURALNET *Net);
void TSetDeSDev(NEURALNET *Net);
void NormalizeDesiredOutput(NEURALNET *Net);
void VSetDeAvg (NEURALNET *Net);
void VSetDeSDev(NEURALNET *Net);
void NormalizeDesiredVSetOutput(NEURALNET *Net);
void QuitPreProcedure(NEURALNET *Net, char *MSE2String);

/***************************************************************************
*
* GLOBAL VARIABLES:
*
****************************************************************************/

#define NUM_STORED_SCREEN_COORD 5
int XPos[NUM_STORED_SCREEN_COORD]; /* general purpose text screen X coordinates array */
int YPos[NUM_STORED_SCREEN_COORD]; /* general purpose text screen Y coordinates array */

/***************************************************************************
*
* MAIN():
*
****************************************************************************/
int main(int argc, char *argv[])
{

NEURALNET Net;

PARFR Pfr;
unsigned Index;
char Type;


float *FTmpPtr;
unsigned i;

unsigned nrows,ncols,ncols_;


/* parameter file lines specification */
unsigned AlphaLines[NALPHALINES]={0,1,2,3,4}; /* see input parameter file */
unsigned FloatLines[NFLOATLINES]={5,6,7,8,9,10,11,12,13};/* see input parameter file */


/* Test command line */
if(argc!=2)Help();

/* store input parameter file name */
Net.ParameterFile=argv[1];


/* Allocate memory to disk I/O speed-up buffer */
Net.buffer=(char *)malloc(BUFSIZE*sizeof(char));
if(!Net.buffer)Quit(MEMORY_ERR,"");


/* Read input parameter file: */

/* Initialize  ParamFileReader */
InitParamFileReader(&Pfr,Net.ParameterFile,FloatLines,AlphaLines,NFLOATLINES,NALPHALINES);


/* Identify the parameter/vector at the i_th line of the parameter file,
 i=0,1,...,NALPHALINES+NFLOATLINES-1 and assign it to the corresponding
 variable. */
ParamStorageID(&Pfr,0,&Index,&Type);
Net.TDIF=Pfr.AlphaDat[Index];       /* 0 Training data input file (TDIF) */
ParamStorageID(&Pfr,1,&Index,&Type);
Net.VDIF=Pfr.AlphaDat[Index];       /* 1  Validation data input file (VDIF) */
ParamStorageID(&Pfr,2,&Index,&Type);
Net.FMLPSF=Pfr.AlphaDat[Index];     /* 2 Final MLP specification output file */
ParamStorageID(&Pfr,3,&Index,&Type);
Net.MSEOF=Pfr.AlphaDat[Index];      /* 3  MLP output_to_TDIF_&_VDIF Mean Square Error (MSE) output file */
ParamStorageID(&Pfr,4,&Index,&Type);
Net.MLPO2VOF=Pfr.AlphaDat[Index];   /* 4  MLP output_to_VDIF output file */
ParamStorageID(&Pfr,5,&Index,&Type);
Net.MSEValStopCrit=Pfr.FloatDat[Index][0]; /* 5  Validation MSE (VMSE) value Stopping Criterion (StC) */
ParamStorageID(&Pfr,6,&Index,&Type);
Net.MSEChRStopCrit=Pfr.FloatDat[Index][0]; /* 6  Absolute Percent VMSE Changing Rate StC (APMChRStC)  */
ParamStorageID(&Pfr,7,&Index,&Type);
Net.SmMSEChNEp2C=Pfr.FloatDat[Index][0]; /* 7 The number of consecutive VMSE Changing Rate values below APMChRStC for which the program assumes convergence to a minimum  */
ParamStorageID(&Pfr,8,&Index,&Type);
Net.Alpha=Pfr.FloatDat[Index][0]; /* 8  Momentum Factor [0 to 1] */
ParamStorageID(&Pfr,9,&Index,&Type);
Net.Phy_A=Pfr.FloatDat[Index][0]; /* 9 Activation Function Phy(v)=A*tanh(B*v) Parameter 'A' */
ParamStorageID(&Pfr,10,&Index,&Type);
Net.Phy_B=Pfr.FloatDat[Index][0]; /* 10 Activation Function Phy(v)=A*tanh(B*v) Parameter 'B' */
ParamStorageID(&Pfr,11,&Index,&Type);
Net.SynIRF=Pfr.FloatDat[Index][0]; /* 11 Synapses random initialization range Fan-In Factor */
ParamStorageID(&Pfr,12,&Index,&Type);
FTmpPtr=Pfr.FloatDat[Index]; /* 12 Architecture Vector - first component is the number of input nodes */

/* Assign float parameters to unsigned parameter Top[] */
Net.NumInputNodes=FTmpPtr[0];
FTmpPtr++;

Net.NumLayers=Pfr.VectSize[Index]-1;

Net.Top=(unsigned *)malloc((unsigned)Net.NumLayers*sizeof(unsigned));
if(!Net.Top)Quit(MEMORY_ERR,"");

for(i=0;i<Net.NumLayers;i++)Net.Top[i]=(unsigned)FTmpPtr[i];

ParamStorageID(&Pfr,13,&Index,&Type);


#if(0)
printf("Index of Net.LVect is %u\n",Index);
exit(0);
#endif

Net.LVect=Pfr.FloatDat[Index]; /* 13 Computational layers Learning Rate Vector  */


#if(0)
puts(Net.TDIF);  /* 0 Training data input file (TDIF) */
puts(Net.TDIF);  /* 1  Validation data input file (VDIF) */
puts(Net.FMLPSF);/* 2 Final MLP specification output file */
puts(Net.MSEOF); /* 3  MLP output_to_TDIF_&_VDIF Mean Square Error (MSE) output file */
puts(Net.MLPO2VOF); /* 4  MLP output_to_VDIF output file */
printf("%G\n",Net.MSEValStopCrit); /* 5  Validation MSE (VMSE) value Stopping Criterion (StC) */
printf("%G\n",Net.MSEChRStopCrit); /* 6  Absolute Percent VMSE Changing Rate StC (APMChRStC)  */
printf("%u\n",Net.SmMSEChNEp2C); /* 7 The number of consecutive VMSE Changing Rate values below APMChRStC for which the program assumes convergence to a minimum  */
printf("%G\n",Net.Alpha);/* 8 Momentum Factor [0 to 1] */
printf("%G\n",Net.Phy_A); /* 9 Activation Function Phy(v)=A*tanh(B*v) Parameter 'A' */
printf("%G\n",Net.Phy_B); /* 10 Activation Function Phy(v)=A*tanh(B*v) Parameter 'B' */
printf("%G\n",Net.SynIRF); /* 11 Synapses random initialization range Fan-In Factor */

for(i=0;i<Net.NumLayers;i++)printf("%u\t",Net.Top[i]);printf("\n"); /*  Architecture Vector */

                    /* | the index 8 float parameter
                       V */
for(i=0;i<Pfr.VectSize[8];i++)printf("%f\t",Net.LVect[i]);printf("\n"); /*  Layer Learning Rate Vector */

exit(0);
#endif


/* Check */
if(Pfr.VectSize[Index]!=Net.NumLayers)Quit(LNVECSIZE_ERR,Net.ParameterFile);


/* Initialize the random number generator random() with a random value */
Randomize();


/* Initialize Net */
InitMLP(&Net);


/* Determine the number of columns in training data input matrix: */
ncols=Net.NumInputNodes+Net.Top[(Net.NumLayers)-1];


/* Read Training Set */
Net.TSet=FloatMatReadCheck(Net.TSet=NULL,Net.TDIF,Net.buffer,&nrows,&ncols_,1);

if(ncols!=ncols_)Quit(SIZE_ERR,Net.TDIF);
Net.TSSize=nrows;

/* Normalize desired output set to range [-1,1] */
NormalizeDesiredOutput(&Net);

/* De-averages Training Set */
TSetDeAvg(&Net);

/* Normalize each TSet input vector component by the corresponding
component of the TSet standard deviation vector. */
TSetDeSDev(&Net);

/* Write normalized/de-averaged/de-deviated input set to disk */
#if(0)
FloatMatWrite(Net.TSet,"tset#.txt",Net.buffer,Net.TSSize,ncols);
#endif


if(Net.ValidationMode){

/* Read Validation Set */
Net.VSet=FloatMatReadCheck(Net.VSet=NULL,Net.VDIF,Net.buffer,&nrows,&ncols_,1);

if(ncols!=ncols_)Quit(SIZE_ERR,Net.VDIF);
Net.VSSize=nrows;

/*  Normalize VSet desired output set to range [-1,1] */
NormalizeDesiredVSetOutput(&Net);

/* De-averages Validation Set */
VSetDeAvg(&Net);

/* Normalize each VSet input vector component by the corresponding
component of the TSet standard deviation vector. */
VSetDeSDev(&Net);

/* Write normalized/de-averaged/de-deviated validation set to disk */
#if(0)
FloatMatWrite(Net.VSet,"vset#.txt",Net.buffer,Net.VSSize,ncols);
#endif
}
else{
Net.VSet=Net.TSet;
Net.VSSize=Net.TSSize;
}


/* Show parameters from parameter file */
ShowParameters(&Net);

/* MLP training loop */
MLPTrainer(&Net);

return 0; 
}

/***************************************************************************
* FUNC: void InitMLP(NEURALNET *Net)
*
* DESC: Initialize NEURALNET Net.
****************************************************************************/
void InitMLP(NEURALNET *Net)
{

unsigned i;

/* check validation mode */
if(!strcmp(Net->TDIF,Net->VDIF))Net->ValidationMode=0;
else Net->ValidationMode=1;


/* Alloc TSet average vector */
Net->Avg=(float *)calloc((unsigned)Net->NumInputNodes,sizeof(float));
if(!Net->Avg)Quit(MEMORY_ERR,"");


/* Alloc TSet standard deviation vector */
Net->SDev=(float *)calloc((unsigned)Net->NumInputNodes,sizeof(float));
if(!Net->SDev)Quit(MEMORY_ERR,"");


/* Alloc NumLayer Layers in Net: */
Net->L=(LAYER *)malloc((unsigned)Net->NumLayers*sizeof(LAYER));
if(!Net->L)Quit(MEMORY_ERR," @ 'Layer'");


/* Alloc ErrVect size equal to the numbers of neurons in last Layer: */
Net->ErrVect=(float *)malloc((unsigned)(Net->Top[(Net->NumLayers)-1])*sizeof(float));
if(!Net->ErrVect)Quit(MEMORY_ERR," @ 'ErrVect'");


/* Init Layers Learnig Rate parameters: */
for(i=0;i<Net->NumLayers;i++)Net->L[i].nt=Net->LVect[i];


AllocLayer(Net);
AllocNeuron(Net);

/* Randomly init the MLP synapses weights */
InitWeight(Net);


}


/***************************************************************************
* FUNC: void AllocLayer(NEURALNET *Net)
*
* DESC: Allocate memory for the neurons in each layer.
****************************************************************************/
void AllocLayer(NEURALNET *Net)

{

register unsigned i;


for(i=0;i<Net->NumLayers;i++){

/* alloc Top[i] Neurons in Layer i: */
Net->L[i].N=(NEURON *)malloc((Net->Top[i])*sizeof(NEURON));
if(!(Net->L[i].N))Quit(MEMORY_ERR," @ 'Neuron'");

}


}


/***************************************************************************
* FUNC: void AllocNeuron(NEURALNET *Net)
*
* DESC: Allocate memory for the synapses weight vectors in each neuron.
****************************************************************************/
void AllocNeuron(NEURALNET *Net)

{

register unsigned i,j; /* i indexes Layer, j indexes Neuron in Layer i */


for(i=0;i<Net->NumLayers;i++){
for(j=0;j<Net->Top[i];j++){


if(i==0){

/* if first Layer, alloc so much synapses weights as the amount of Input
Nodes plus one (for bias) */
Net->L[i].N[j].W=(float *)malloc(((Net->NumInputNodes)+1)*sizeof(float));
if(!(Net->L[i].N[j].W))Quit(MEMORY_ERR," @ 'Synapse Weight' - 1st layer");

/* if first Layer, alloc so much previous cycle synapses weights correction
factors as the amount of Input Nodes plus one (for bias) */
Net->L[i].N[j].DW_1=(float *)malloc(((Net->NumInputNodes)+1)*sizeof(float));
if(!(Net->L[i].N[j].DW_1))Quit(MEMORY_ERR," @ 'Previous Cycle Weight Correction'- 1st layer");


//printf("i=%u j=%u size=%u\n",i,j,(Net->Top[i])+1);
}/* if i==0 */


else{
/* if not first Layer, alloc so much synapses weights as the amount of Neurons
 in previous Layer plus one (for bias) */
Net->L[i].N[j].W=(float *)malloc(((Net->Top[i-1])+1)*sizeof(float));
if(!(Net->L[i].N[j].W))Quit(MEMORY_ERR," @ 'Synapse Weight'");

/* if not first Layer, alloc so much previous cycle synapses weights correction
factors as the amount of Neurons in previous Layer plus one (for bias) */
Net->L[i].N[j].DW_1=(float *)malloc(((Net->Top[i-1])+1)*sizeof(float));
if(!(Net->L[i].N[j].DW_1))Quit(MEMORY_ERR," @ 'Previous Cycle Weight Correction'");

//printf("i=%u j=%u size=%u\n",i,j,(Net->Top[i-1])+1);
}/* else */



}/* for j */
}/* for i */


}


/***************************************************************************
* FUNC: float WeightGen(NEURALNET *Net,unsigned FanIn)
*
* DESC: Generate MLP synapses weights random values.
****************************************************************************/
float WeightGen(NEURALNET *Net,unsigned FanIn)

{

return  (((float)rand()/(RAND_MAX/2))-1)*(float)Net->SynIRF/FanIn;

}


/***************************************************************************
* FUNC: void InitWeight(NEURALNET *Net)
*
* DESC: Randomly initialize the MLP synapses weights.
****************************************************************************/
void InitWeight(NEURALNET *Net)
{


register unsigned i; /*  Layer index */
register unsigned j; /*  Neuron index */
register unsigned k; /*  Synapse Weight index NOTE: W[0]=>bias; */


for(i=0;i<Net->NumLayers;i++){
for(j=0;j<Net->Top[i];j++){


if(i==0){


for(k=0;k<=Net->NumInputNodes;k++){  /* <= is due to bias synapse */
/* if first Layer, init synapses weights ramdomly accordingly with neuron
FanIn: */
Net->L[i].N[j].W[k]=WeightGen(Net,Net->NumInputNodes);
/* zeroes previous synaptic weights correction factors vector: */
Net->L[i].N[j].DW_1[k]=0;
}/* for k */


}/* if i==0 */


else{


for(k=0;k<=Net->Top[i-1];k++){  /* <= is due to bias synapse */
/* if not first Layer, init synapses weights ramdomly accordingly with neuron
FanIn (related to the number of neurons in previous Layer): */
Net->L[i].N[j].W[k]=WeightGen(Net,Net->Top[i-1]);
/* zeroes previous synaptic weights correction factors vector: */
Net->L[i].N[j].DW_1[k]=0;
}/* for k */


}/* else */




}/* for j */
}/* for i */



}


/***************************************************************************
* FUNC: void RowShuffle(float **Mat, unsigned row)
*
* DESC: Randomly shuffles the rows of matrix Mat.
****************************************************************************/
void RowShuffle(float **Mat, unsigned row)
{

register unsigned i;
unsigned r1;
unsigned r2;

float *trow;


for(i=0;i<row;i++)
{

r1=rand()%row;
r2=rand()%row;


trow=Mat[r1];
Mat[r1]=Mat[r2];
Mat[r2]=trow;

}



}



/***************************************************************************
* FUNC: void ShowParameters(NEURALNET *Net)
*
* DESC: Show MLP parameters on screen.
****************************************************************************/
void ShowParameters(NEURALNET *Net)

{

int TmpY;
char TmpStr[128];
register unsigned i;


system("cls");

strcpy(TmpStr,"Multilayer Perceptron Training/Validation Program V1.2.d");
gotoxy((80-strlen(TmpStr))/2,2);textcolor(YELLOW);cprintf(TmpStr);

gotoxy(1,4);
textcolor(LIGHTCYAN);cprintf(" Training Data Input File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->TDIF));
textcolor(LIGHTCYAN);cprintf(" Validation Data Input File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->VDIF));
textcolor(LIGHTCYAN);cprintf(" Final MLP Specification Output File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->FMLPSF));
textcolor(LIGHTCYAN);cprintf(" Training MSE & Validation MSE (TMSE & VMSE) Output File: ");textcolor(LIGHTGRAY);cprintf("%s\n\r",strupr(Net->MSEOF));
textcolor(LIGHTCYAN);cprintf(" Validation MSE (VMSE) Value Stopping Criterion: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->MSEValStopCrit);
textcolor(LIGHTCYAN);cprintf(" VMSE Changing Rate Stopping Criterion (VMSEChRSC): ");textcolor(LIGHTGRAY);cprintf("%-8.3G%c\n\r",-1*Net->MSEChRStopCrit,'%');/* -Net->MSEChRStopCrit because it's a diminishing changing rate */
textcolor(LIGHTCYAN);cprintf(" Number of Consecutive VMSEChRSC Attendance Stopping Criterion: ");textcolor(LIGHTGRAY);cprintf("%u\n\r",Net->SmMSEChNEp2C);
textcolor(LIGHTCYAN);cprintf(" Momentum Factor: ");textcolor(LIGHTGRAY);cprintf("%-8.3G\n\r",Net->Alpha);
textcolor(LIGHTCYAN);cprintf(" Activation Function Phy(v)=A*tanh(B*v) Parameter 'A': ");textcolor(LIGHTGRAY);cprintf("%-8.4G\n\r",Net->Phy_A);
textcolor(LIGHTCYAN);cprintf(" Activation Function Phy(v)=A*tanh(B*v) Parameter 'B': ");textcolor(LIGHTGRAY);cprintf("%-8.4G\n\r",Net->Phy_B);
textcolor(LIGHTCYAN);cprintf(" Synapses Weight Initialization Range Factor: ");textcolor(LIGHTGRAY);cprintf("%-5.2f\n\r",Net->SynIRF);
textcolor(LIGHTCYAN);cprintf("\n\r Architecture: ");textcolor(WHITE);cprintf("%u ",Net->NumInputNodes);for(i=0;i<Net->NumLayers;i++)cprintf("%u ",Net->Top[i]);
textcolor(LIGHTCYAN);cprintf("\n\r Learning Rate: ");textcolor(WHITE);for(i=0;i<Net->NumLayers;i++)cprintf("%5.3G ",Net->LVect[i]);
textcolor(LIGHTCYAN);cprintf("\n\r Training Set Size: ");textcolor(WHITE);cprintf("%u\n\r",Net->TSSize);

TmpY=wherey()+1;

textcolor(LIGHTMAGENTA);

gotoxy(1,TmpY);cprintf(" Epoch:");XPos[0]=wherex();YPos[0]=wherey();
               
gotoxy(1,TmpY+1);cprintf(" TMSE:");XPos[2]=wherex();YPos[2]=wherey();

gotoxy(1,TmpY+2);cprintf(" VMSE:");XPos[3]=wherex();YPos[3]=wherey();

gotoxy(32,TmpY+2);cprintf("VMSE Change Rate:");XPos[4]=wherex();YPos[4]=wherey();

gotoxy(32,TmpY+1);cprintf("VMSE Change: ");XPos[1]=wherex();YPos[1]=wherey();

}



/***************************************************************************
* FUNC: float ForwardPass(NEURALNET *Net, float *InputData)
*
* DESC: ForwardPass() performs the forward pass training procedure. Return the
*       instantaneous sum of the squared errors for the present pass.
****************************************************************************/
float ForwardPass(NEURALNET *Net, float *InputData)
{

float sqe;           /* instantaneous squared error */
register unsigned i; /*  Layer index */
register unsigned j; /*  Neuron index */
register unsigned k; /*  Synapse Weight (and Input Signal) index
			 NOTE: W[0]=>bias Weight */



/*###################################
Compute Neurons Function Signal 'y':
####################################*/


/*-------------------------------
Neurons in the First Layer - i=0:
--------------------------------*/


for(j=0;j<Net->Top[0];j++){


Net->L[0].N[j].v=BIAS*(Net->L[0].N[j].W[0]); /* activation level from bias */


for(k=1;k<=Net->NumInputNodes;k++){  /* k=1 and <= is due to bias synapse */

Net->L[0].N[j].v+=(Net->L[0].N[j].W[k])*(InputData[k-1]); /* neuron summing junction */

}/* for k */


Net->L[0].N[j].v=Phy(Net,&Net->L[0].N[j].v); /* transforms neuron activation level
					 'v' in neuron function signal 'y' by
					  application of activation function
					  Phy(v)) */
} /* for j */



/*-----------------------------------
Neurons not in the first Layer - i>0:
-------------------------------------*/

for(i=1;i<Net->NumLayers;i++){
for(j=0;j<Net->Top[i];j++){


Net->L[i].N[j].v=BIAS*(Net->L[i].N[j].W[0]); /* activation level from bias */



for(k=1;k<=Net->Top[i-1];k++){  /* k=1 and <= is due to bias synapse */


/* Note that the numbers of neurons in the left previous layer (i-1) equals the
 number of synapses (k indexed) of neuron j at the present layer (i) minus 1 */
Net->L[i].N[j].v+=(Net->L[i].N[j].W[k])*(Net->L[i-1].N[k-1].v); /* neuron summing junction */

}/* for k */


Net->L[i].N[j].v=Phy(Net,&Net->L[i].N[j].v); /* transforms neuron activation level
					 'v' in neuron function signal 'y' by
					  application of activation function
					  Phy(v)) */


}/* for j */
}/* for i */



/*########################################
Determine the Error Vector from desired data
in Input Data and data in Output Layer:
#########################################*/

for(j=0;j<(Net->Top[(Net->NumLayers)-1]);j++){

Net->ErrVect[j]= InputData[j+(Net->NumInputNodes)] - (Net->L[(Net->NumLayers)-1].N[j].v);

}


/*###########################################
Compute instantaneous sum of squared errors:
#############################################*/


sqe=0;
for(j=0;j<(Net->Top[(Net->NumLayers)-1]);j++){

sqe+=SQR(Net->ErrVect[j]);

}

return sqe/2;

}




/***************************************************************************
* FUNC: void BackwardPass(NEURALNET *Net, float *InputData)
*
* DESC: BackwardPass() performs the Back Propagation Algorithm.
****************************************************************************/
void BackwardPass(NEURALNET *Net, float *InputData)
{

int i; /*  Layer index */
register unsigned j; /*  Neuron index */
register unsigned k; /*  Right Layer neuron index */

float sigma;
float DW;


/*##############################
Compute Neurons Local Grads:
###############################*/


/*-------------------------------------------------
Neurons in the Output Layer - i=(Net->NumLayers)-1:
--------------------------------------------------*/

for(j=0;j<Net->Top[(Net->NumLayers)-1];j++){

Net->L[(Net->NumLayers)-1].N[j].d=Net->ErrVect[j]*DPhy(Net,Net->L[(Net->NumLayers)-1].N[j].v);

} /* for j */



/*-----------------------------------------------------
Neurons Not in the Output Layer - i<(Net->NumLayers)-1:
-------------------------------------------------------*/

for(i=(Net->NumLayers)-2;i>=0;i--){ /* starts from left to the right */
for(j=0;j<Net->Top[i];j++){


sigma=0;

/* Sum all neurons Local Grads d in the layer to the right multiplied by its
synapses Wkj which connects to neuron j in the present layer (to the left)
- see Equation (4.24) Haykin */
for(k=0;k<Net->Top[i+1];k++){

sigma+=(Net->L[i+1].N[k].d)*(Net->L[i+1].N[k].W[j]);

}/* for k */


/* Compute the Local Grad d of Neuron j in the present Layer
 - see Equation (4.24) Haykin */
Net->L[i].N[j].d=DPhy(Net,Net->L[i].N[j].v)*sigma;


}/* for j */
}/* for i */


/*########################
Update Synapses Weights:
#########################*/


/*-------------------------------
Neurons in the First Layer - i=0:
--------------------------------*/


for(j=0;j<Net->Top[0];j++){ /* j indexes Neuron */


/* Procedure for bias synapses: */

/* compute DW (delta weight) for bias synapse - See Equation (4.39) Haykin */
DW=Net->Alpha*(Net->L[0].N[j].DW_1[0])+(Net->L[0].nt)*(Net->L[0].N[j].d)*BIAS;

/* Update weight for bias synapse */
Net->L[0].N[j].W[0]+=DW;

/* Update DW_1 (previous delta weight) for bias synapse */
Net->L[0].N[j].DW_1[0]=DW;




/* Procedure for NOT bias synapses: */

for(k=1;k<=Net->NumInputNodes;k++){  /* k=1 and <= is due to bias synapse */

/* Compute DW (delta weight) for synapse k - See Equation (4.39) Haykin */
DW=Net->Alpha*(Net->L[0].N[j].DW_1[k])+(Net->L[0].nt)*(Net->L[0].N[j].d)*(InputData[k-1]);

/* Update weight for synapse k */
Net->L[0].N[j].W[k]+=DW;

/* Update DW_1 (previous delta weight) for synapse k */
Net->L[0].N[j].DW_1[k]=DW;

}/* for k */




}/* for j */

/*-----------------------------------
Neurons not in the first Layer - i>0:
-------------------------------------*/

for(i=1;i<Net->NumLayers;i++){ /* i indexes Layer */
for(j=0;j<Net->Top[i];j++){    /* j indexes Neuron */



/* Procedure for bias synapses: */

/* Compute DW (delta weight) for bias synapse. See Equation (4.39) Haykin */
DW=Net->Alpha*(Net->L[i].N[j].DW_1[0])+(Net->L[i].nt)*(Net->L[i].N[j].d)*BIAS;

/* Update weight for bias synapse */
Net->L[i].N[j].W[0]+=DW;

/* Update DW_1 (previous delta weight) for bias synapse */
Net->L[i].N[j].DW_1[0]=DW;




/* Procedure for NOT bias synapses: */

for(k=1;k<=Net->Top[i-1];k++){  /* k=1 and <= is due to bias synapse */


/* Compute DW (delta weight) for synapse k. See Equation (4.39) Haykin. */
DW=Net->Alpha*(Net->L[i].N[j].DW_1[k])+(Net->L[i].nt)*(Net->L[i].N[j].d)*(Net->L[i-1].N[k-1].v);

/* Update weight for synapse k */
Net->L[i].N[j].W[k]+=DW;

/* Update DW_1 (previous delta weight) for synapse k */
Net->L[i].N[j].DW_1[k]=DW;


}/* for k */



}/* for j */
}/* for i */



}


/***************************************************************************
* FUNC: inline float Phy(NEURALNET *Net,float *v)
*
* DESC: Phy() is the hyperbolic tangent activation function.
*
* NOTE: 2A/(1+e^-2Bv)-A=A(1-e^-2Bv)/(1+e^-2Bv)=Atanh(Bv)
****************************************************************************/
inline float Phy(NEURALNET *Net,float *v)

{
static float u;

u=-2*Net->Phy_B*(*v); /* u=-2Bv */

if(u>UMAX)u=UMAX; /* avoid  float overflow */

return (2*Net->Phy_A/(1+exp(u))-Net->Phy_A);  /* Atanh(Bv) */

}


/***************************************************************************
* FUNC: inline float DPhy(NEURALNET *Net,float y)
*
* DESC: DPhy() returns the derivative of Phy().
****************************************************************************/
inline float DPhy(NEURALNET *Net,float y)

{

return (Net->Phy_B/Net->Phy_A)*(Net->Phy_A-y)*(Net->Phy_A+y);

}



/***************************************************************************
* FUNC: void Help(void)
*
* DESC: Show the help menu on screen.
****************************************************************************/

void Help(void)
{

char TmpStr[128];

system("cls");
strcpy(TmpStr,"Multilayer Perceptron Training/Validation Program V1.2.d");
gotoxy((80-strlen(TmpStr))/2,2);textcolor(YELLOW);cprintf(TmpStr);

strcpy(TmpStr,"By Fernando & Cristina de Castro - September 2001");
gotoxy((80-strlen(TmpStr))/2,4);textcolor(LIGHTBLUE);cprintf(TmpStr);

gotoxy(1,7);textcolor(LIGHTCYAN);cputs("Use: MLPTrv ParameterFile\n");

textcolor(LIGHTGRAY);
exit(1);

}


/***************************************************************************
* FUNC: char ChangeParameterMenu(NEURALNET *Net)
*
* DESC: Change MLP parameters "on fly". Return the performed change
*       operation.
****************************************************************************/
char ChangeParameterMenu(NEURALNET *Net)
{

#define CGETS_NCH 80 /* max number of chars read by cgets() */

register unsigned i;
int xpos[7],ypos[7];
char TmpStr[128];
char op,op_1;
static char str[CGETS_NCH+3];
char *p;

str[0]=CGETS_NCH+1;  /* There's space for CGETS_NCH chars plus the NULL terminator */
op=op_1=0;
getch();


while(1){

system("cls");

strcpy(TmpStr,"MLP Neural Network Parameters Status:");
gotoxy((80-strlen(TmpStr))/2,1);textcolor(YELLOW);cprintf(TmpStr);
gotoxy(1,2);
textcolor(LIGHTCYAN);cprintf(" MLP Specification File: ");textcolor(LIGHTGRAY);xpos[0]=wherex();ypos[0]=wherey();cprintf("%s\n\r",strupr(Net->FMLPSF));
textcolor(LIGHTCYAN);cprintf(" Validation MSE (VMSE) Value Stopping Criterion: ");textcolor(LIGHTGRAY);xpos[1]=wherex();ypos[1]=wherey();cprintf("%-8.3G\n\r",Net->MSEValStopCrit);
textcolor(LIGHTCYAN);cprintf(" VMSE Changing Rate Stopping Criterion: ");textcolor(LIGHTGRAY);xpos[2]=wherex();ypos[2]=wherey();cprintf("%-8.3G%c\n\r",-1*Net->MSEChRStopCrit,'%');/* -Net->MSEChRStopCrit because it's a diminishing changing rate */
textcolor(LIGHTCYAN);cprintf(" Momentum Factor: ");textcolor(LIGHTGRAY);xpos[3]=wherex();ypos[3]=wherey();cprintf("%-8.3G\n\r",Net->Alpha);
textcolor(LIGHTCYAN);cprintf(" Activation Function Phy(v)=A*tanh(B*v) Parameter 'A': ");textcolor(LIGHTGRAY);xpos[4]=wherex();ypos[4]=wherey();cprintf("%-8.4G\n\r",Net->Phy_A);
textcolor(LIGHTCYAN);cprintf(" Activation Function Phy(v)=A*tanh(B*v) Parameter 'B': ");textcolor(LIGHTGRAY);xpos[5]=wherex();ypos[5]=wherey();cprintf("%-8.4G\n\r",Net->Phy_B);
textcolor(LIGHTCYAN);cprintf(" Architecture: ");textcolor(WHITE);cprintf("%u ",Net->NumInputNodes);for(i=0;i<Net->NumLayers;i++)cprintf("%u ",Net->Top[i]);
textcolor(LIGHTCYAN);cprintf("\n\r Learning Rate: ");textcolor(WHITE);xpos[6]=wherex()+1;ypos[6]=wherey();for(i=0;i<Net->NumLayers;i++)cprintf("%5.3G ",Net->LVect[i]);

strcpy(TmpStr,"Change Menu:");
gotoxy((80-strlen(TmpStr))/2,wherey()+1);textcolor(YELLOW);cprintf(TmpStr);cprintf("%\n\r");
textcolor(LIGHTCYAN);cprintf(" MS");textcolor(LIGHTRED);cprintf("E");textcolor(LIGHTCYAN);cprintf(" Value Stopping Criterion\n\r");
cprintf(" MSE ");textcolor(LIGHTRED);cprintf("C");textcolor(LIGHTCYAN);cprintf("hanging Rate Stopping Criterion\n\r");
textcolor(LIGHTRED);cprintf(" M");textcolor(LIGHTCYAN);cprintf("omentum Factor\r\n");
cprintf(" Activation Func. Phy(v)=");textcolor(LIGHTRED);cprintf("A");textcolor(LIGHTCYAN);cprintf("*tanh(B*v) Parameter 'A'\n\r");
cprintf(" Activation Func. Phy(v)=A*tanh(");textcolor(LIGHTRED);cprintf("B");textcolor(LIGHTCYAN);cprintf("*v) Parameter 'B'\n\r");
textcolor(LIGHTRED);cprintf(" L");textcolor(LIGHTCYAN);cprintf("earning Rate\r\n");
textcolor(LIGHTRED);cprintf(" R");textcolor(LIGHTCYAN);cprintf("ead Neural Network from file\r\n");
textcolor(LIGHTRED);cprintf(" W");textcolor(LIGHTCYAN);cprintf("rite Neural Network to file\r\n");
textcolor(LIGHTRED);cprintf(" G");textcolor(LIGHTCYAN);cprintf("o ahead\r\n");
textcolor(LIGHTRED);cprintf(" Q");textcolor(LIGHTCYAN);cprintf("uit and save Neural Network\r\n");
textcolor(LIGHTCYAN);cprintf(" Ab");textcolor(LIGHTRED);cprintf("o");textcolor(LIGHTCYAN);cprintf("rt and don't save Neural Network");


gotoxy(1,wherey()+1);textcolor(LIGHTRED);cprintf(" Your Change Menu option?");
cscanf("%c",&op);
op=toupper(op);

textcolor(YELLOW);

switch (op) {
  case 'R':
  gotoxy(xpos[0],ypos[0]);
  textbackground(LIGHTGRAY);
  textcolor(LIGHTRED);
  for(i=0;i<strlen(Net->FMLPSF);i++)cprintf("%c",' ');
  gotoxy(xpos[0],ypos[0]);
  p = cgets(str);
  NetRead(Net,p);
  #if(0)
  NetWrite(Net,"netwrt.txt");
  #endif
  break;
  case 'W':
  gotoxy(xpos[0],ypos[0]);
  textbackground(LIGHTGRAY);
  textcolor(LIGHTRED);
  for(i=0;i<strlen(Net->FMLPSF);i++)cprintf("%c",' ');
  gotoxy(xpos[0],ypos[0]);
  p = cgets(str);
  NetWrite(Net,p);
  break;
  case 'E':
  gotoxy(xpos[1],ypos[1]);
  cscanf("%f",&Net->MSEValStopCrit);
  break;
  case 'C':
  gotoxy(xpos[2],ypos[2]);
  cscanf("%f",&Net->MSEChRStopCrit);
  Net->MSEChRStopCrit=fabs(Net->MSEChRStopCrit);
  break;
  case 'M':
  gotoxy(xpos[3],ypos[3]);
  cscanf("%f",&Net->Alpha);
  break;
  case 'A':
  gotoxy(xpos[4],ypos[4]);
  cscanf("%f",&Net->Phy_A);
  break;
  case 'B':
  gotoxy(xpos[5],ypos[5]);
  cscanf("%f",&Net->Phy_B);
  break;
  case 'L':
  gotoxy(xpos[6],ypos[6]);
  for(i=0;i<Net->NumLayers;i++)
  cscanf("%f",&Net->LVect[i]);
  break;
  case 'G':
  return op_1;
  case 'Q':
  QuitPreProcedure(Net,TmpStr);
  Quit(USER_INT,Net->FMLPSF);
  break;
  case 'O':
  Quit(ABORT_INT,"");
  break;
  default:
  continue;

}

op_1=op;

}/* while(1) */


}


/***************************************************************************
* FUNC: void  Randomize(void)
*
* DESC: Randomize random generator seed.
****************************************************************************/
void Randomize(void)
{
srand((unsigned) time(NULL));
}


/*****************************************************************************
* FUNC: float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  float matrix.
*****************************************************************************/
float **FloatMatCAlloc(unsigned NRows,unsigned NCols)
{

float  *x;
float  **Mx;
register unsigned i;


x = (float *)calloc((unsigned long)(NRows*NCols),sizeof(float));
if(!x)Quit(MEMORY_ERR,"");


Mx=(float  **)calloc((unsigned long)NRows,sizeof(float *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(float *)&x[i*NCols];

}

return Mx;
}



/*****************************************************************************
* FUNC: void FloatMatFree(float **M)
*
* DESC: Free memory for the float matrix M.
*****************************************************************************/
void FloatMatFree(float **M)
{

free(M[0]);
free(M);

}


/*****************************************************************************
* FUNC: void Quit(EVENT Evt, char *Message)
*
* DESC: Print Event message and exit to shell.
******************************************************************************/
void Quit(EVENT Evt, char *Message)
{

static char *Event[] = {

/*  NO_EVENT */       " ",
/*  MEMORY_ERR */     "Not enough memory",
/*  READ_OPEN_ERR */  "I can't find the file",
/*  READ_ERR */       "Error reading file",
/*  WRITE_OPEN_ERR */ "I can't open the file",
/*  WRITE_ERR */      "I can't write the file",
/*  BUFFER_ERR */     "I can't attach buffer to file",
/*  SIZE_ERR */       "Specified size doesn't match data size in file",
/*  MAXIT_ERR */      "Exceeded max number of iterations in function",
/*  LOGIC_ERR */      "Logic error in function",
/*  READ_NNET_ERR */  "Incompatible neural network in file",
/*  MTXDIM_ERR */     "Invalid matrix dimensions in file",
/*  PARF_ERR  */      "Wrong parameter file",
/*  GFTD_ERR  */      "I can't get time & date info from file",
/*  MAXFLOAT_READ_ERR */"Exceeded float maximum absolute value reading file",
/*---------*/
/*  USER_INT */	      "Training interrupted by user! Neural Network saved to file",
/*  SMALL_MSECHG_INT */"Training terminated by small enough VMSE Change Rate with VMSE =",
/*  SAT_INT */        "Training terminated due to neurons saturation at VMSE =", 
/*  SMALL_MSE_INT */  "Training terminated by small enough VMSE =",
/*  EARLY_EOF_ERR */  "Found premature EOF in file ",
/*  HEADER_SIZE_ERR */"Found impossible number of samples\nin header of file ",
/*  ABORT_INT */      "Training aborted! Neural Network not saved",
/*  LNVECSIZE_ERR */  "Wrong layer Learning Rate Vector size in file"

};


        
	system("cls");
	textcolor(LIGHTGRAY);

	if (Evt != NO_EVENT) {


		if(Message[0]){
		fprintf (stderr,"%s %s!\n", Event[Evt], Message);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}
		else{
		fprintf (stderr,"%s!\n", Event[Evt]);
                #if(0)
		putch(BEL);
                #endif
		exit (Evt);
		}

	}
	exit (0);
}


/*****************************************************************************
* FUNC: unsigned long GetFileNumLines(FILE *Fp)
*
* DESC: Return the number of lines in the text file pointed by Fp.
******************************************************************************/
unsigned long GetFileNumLines(FILE *Fp)
{

  char line[MAXLINESIZE];
  register unsigned long Count;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user have typed it */


  rewind(Fp);
  return Count;

}



/*****************************************************************************
* FUNC: long GetFileNumSamples(FILE *Fp)
*
* DESC: Return the number of numeric samples in the text file pointed by Fp.
*       Return -1 on read error.
*****************************************************************************/
long GetFileNumSamples(FILE *Fp)
{

register long Count;
float x;
int test;

Count=0;

while((test=fscanf(Fp,"%f",&x))!=-1){
if(test!=1)return -1;
Count++;
}

rewind(Fp);
return Count;

}


/*****************************************************************************
* FUNC: float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
*
* DESC: Read an ascii float matrix from 'file' and store it in matrix
*       M[NRows][NCols]. If AllocFlag==1 the function allocates memory
*       for M and the arguments *NRows and *NCols identify the size of M.
*
*       FloatMatReadCheck() checks the float compatibility of the read data.
*****************************************************************************/
float  **FloatMatReadCheck(float **M,char *file,char *buffer,unsigned *NRows,unsigned *NCols, char AllocFlag)
{

register unsigned i,j;
FILE *in;
long NumSamples;
unsigned long NumLines;
int Test;
double Value;

if((in=fopen(file,"rt"))==NULL)Quit(READ_OPEN_ERR,file);
if (setvbuf(in, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,file);

NumSamples=GetFileNumSamples(in);
if(NumSamples==-1)Quit(READ_ERR,file);
NumLines=GetFileNumLines(in);

if(NumSamples%NumLines)Quit(MTXDIM_ERR,file); /* check valid matrix dimensions */

*NRows=NumLines;
*NCols=NumSamples/NumLines;

if(AllocFlag==1)M=FloatMatCAlloc(*NRows,*NCols);

for(i=0;i<*NRows;i++){
for(j=0;j<*NCols;j++){
Test=fscanf(in,"%lf",&Value);
if(Test!=1)Quit(READ_ERR,file);
if(fabs(Value)>MAXFLOAT)Quit(MAXFLOAT_READ_ERR,file);
M[i][j]=Value;
}
}

fclose(in);

return M;
}




/*****************************************************************************
* FUNC: void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols)
*
* DESC: Write a float NRows x NCols matrix to an ascii file. If file=NULL, it
*       writes to stdout.
*****************************************************************************/
void FloatMatWrite(float **M,char *file,char *buffer,unsigned NRows,unsigned NCols)
{

register unsigned i,j;
FILE *out;
int test;

if(file!=NULL){
if((out=fopen(file,"wt"))==NULL)Quit(WRITE_OPEN_ERR,file);
if (setvbuf(out, buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,file);
}
else out=stdout;


for(i=0;i<NRows;i++){
for(j=0;j<NCols;j++){

fprintf(out,"%8.6G\t",M[i][j]);

}

test=fprintf(out,"\n");
if(test!=sizeof(char))Quit(WRITE_ERR,file);

}

if(file!=NULL)fclose(out);


}


/*****************************************************************************
* FUNC: void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)
*
* DESC: Init parameter file reader.
*****************************************************************************/
void InitParamFileReader(PARFR *Pfr,char *ParameterFile,unsigned *FloatLines,unsigned *AlphaLines, unsigned NumFloatLines, unsigned NumAlphaLines)

{

register unsigned i;

Pfr->ParameterFile=ParameterFile;
Pfr->FloatLines=FloatLines;
Pfr->AlphaLines=AlphaLines;
Pfr->NumFloatLines=NumFloatLines;
Pfr->NumAlphaLines=NumAlphaLines;


/* Alloc memory for disk I/O speed-up buffers */
Pfr->BufferI=(char *)calloc(SMALLBUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, SMALLBUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);

/* Alloc memory for Sheet */
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength);

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);

/* Alloc memory for float FloatDat row pointers */
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){
Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}


/*****************************************************************************
* FUNC: void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
*
* DESC: StorageIndex return the index in respective (FloatDat/AlphaDat)
*       storage sheet that corresponds to line LineNumber.
*
*       Type returns the respective data type: 'f'->float 'a'->alpha
*****************************************************************************/
void ParamStorageID(PARFR *Pfr, unsigned LineNumber, unsigned *StorageIndex, char *Type)
{

register unsigned i;
int p=-1;
char ParType=0;

//printf("LineNumber=%u\n",LineNumber);

/* scan FloatLines index array */
for(i=0;i<Pfr->NumFloatLines;i++){
if(Pfr->FloatLines[i]==LineNumber){p=i;ParType='f';break;}
}

/* if not found, scan AlphaLines index array */
if(p==-1){
for(i=0;i<Pfr->NumAlphaLines;i++){
if(Pfr->AlphaLines[i]==LineNumber){p=i;ParType='a';break;}
}

}

/* not found */
if(p==-1){fprintf(stderr,"No such line number in parameter file %s!\n",Pfr->ParameterFile);putch(BEL);exit(1);}

/* found */
*StorageIndex=p;
*Type=ParType;

}


/*****************************************************************************
* FUNC: unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
*
* DESC: Return number of lines of the text sheet in file pointed by Fp.
*       MaxLength return the line max length, including the '\0'.
******************************************************************************/
unsigned long GetSheetNumLines(FILE *Fp, unsigned *MaxLength)
{

  char line[MAXLINESIZE];
  register unsigned long Count;
  unsigned Length;

  *MaxLength=0;


  Count=0;
  while (fgets(line,MAXLINESIZE-2,Fp)){
  if((Length=strlen(line))>*MaxLength)*MaxLength=Length;
  Count++;
  }

  if(strlen(line)==1)Count--;/* to discount the last \n in file
                              for the case the user has typed it */

  *MaxLength= (*MaxLength)+1; /* to handle the '\0' added by fgets() */

  rewind(Fp);
  return Count;


}


/*****************************************************************************
* FUNC: char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
*
* DESC: Allocate memory for a NRows x NCols  char matrix.
*****************************************************************************/
char **CharMatCAlloc(unsigned long NRows,unsigned NCols)
{
char  *x;
char  **Mx;
register unsigned  i;

x = (char  *)calloc((unsigned long)(NRows*NCols),sizeof(char));
if(!x)Quit(MEMORY_ERR,"");

Mx=(char  **)calloc((unsigned long)NRows,sizeof(char *));
if(!Mx)Quit(MEMORY_ERR,"");


for(i=0;i<NRows;i++){

Mx[i]=(char *)&x[i*NCols];

}

return Mx;
}


/*****************************************************************************
* FUNC: unsigned NumberOfFloatsInLine(char *Line)
*
* DESC: Return the number of floats in Line, starting from the beginning.
*       Upon finding a non-readable character, the function ends the count.
*****************************************************************************/
unsigned NumberOfFloatsInLine(char *Line)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

#if(0)
puts(Line);
#endif

while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

#if(0)
fprintf(stdout,"Val[%u]=%8.6lg%c\n",Count,Val,NonReadable[0]);
#endif

Line+=strlen(Str)+1;
Count++;
}

#if(0)
fprintf(stdout,"Line has %u floats.\n",Count);
#endif

return Count;
}


/*****************************************************************************
* FUNC: float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
*
* DESC:  Convert the floats in Line into a float vector Vector, starting from the
*        beginning.
*
*        Upon finding a non-readable character, the function ends the count.
*        VectSize returns the size of Vector. If Vector==NULL, then memory
*        is allocated for the float vector Vector got from Line.
*****************************************************************************/
float *GetFloatVectFromLine(char *Line, unsigned *VectSize, float *Vector)
{
static char Str[MAXLINESIZE];
static char Buf[MAXLINESIZE];

char *NonReadable;
double Val;
unsigned Count=0;

*VectSize=NumberOfFloatsInLine(Line);
if(*VectSize==0)return NULL;

if(Vector==NULL){
Vector=(float *)calloc(*VectSize,sizeof(float));
if(!Vector)Quit(MEMORY_ERR,"");
}


while(1){
sprintf(Buf,"%s",Line);
if(Buf[0]=='\t'){Line++;continue;} /* found a tab */
if(Buf[0]==' '){Line++;continue;}  /* found a space */
if(!Buf[0])break; /* found a '\0' */
if(Buf[0]=='\n')break; /* found a '\n' */
sscanf(Buf,"%s",Str);
Val= strtod(Str, &NonReadable);

#if(0)
if(NonReadable[0]&&NonReadable[0]!='i')break; /* Check non-readable except 'i'-> to handle complex numbers */
#else
if(NonReadable[0])break; /* Check non-readable */
#endif

Vector[Count]=AULIM(Val); /* check float absolute upper limit */
if(Vector[Count]!=(float)Val){putch(BEL);fprintf(stderr,"Exceeded float maximum absolute value reading parameter from disk!");exit(1);}
Line+=strlen(Str)+1;
Count++;
}


return Vector;
}



/*****************************************************************************
* FUNC: void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
*
* DESC: Reads automatically a parameter file.If AllocFlag!=0, memory is allocated.
*****************************************************************************/
void Automatic_ParamFileReader(PARFR *Pfr,char *ParameterFile, char AllocFlag)
{

#define SIZE_GUARD 6 /* a guard size for Pfr->Sheet line length to avoid eventual diferent float read from ParameterFile */

register unsigned i,j,k;

Pfr->AllocFlag=AllocFlag;
Pfr->ParameterFile=ParameterFile;


Pfr->NumFloatLines=0;
Pfr->NumAlphaLines=0;


/* Alloc memory for disk I/O speed-up buffers */
if(AllocFlag){
Pfr->BufferI=(char *)calloc(BUFSIZE,sizeof(char));
if(!Pfr->BufferI)Quit(MEMORY_ERR,"");
}


/* Open parameter file */
if((Pfr->IFp=fopen(Pfr->ParameterFile,"rt"))==NULL)Quit(READ_OPEN_ERR,Pfr->ParameterFile);
if (setvbuf(Pfr->IFp, Pfr->BufferI, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Pfr->ParameterFile);

/* Get the total number of lines and the max line length */
Pfr->NumLines=GetSheetNumLines(Pfr->IFp,&Pfr->LineMaxLength);


/* Alloc memory for Sheet */
if(AllocFlag){
Pfr->Sheet=CharMatCAlloc(Pfr->NumLines,Pfr->LineMaxLength+SIZE_GUARD);
}

/* Read sheet and store it into Sheet */
for(i=0;i<Pfr->NumLines;i++)fgets(Pfr->Sheet[i],Pfr->LineMaxLength,Pfr->IFp);

/* close parameter file */
fclose(Pfr->IFp);


/* get NumFloatLines and NumAlphaLines from char Sheet */
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->NumAlphaLines++;
else Pfr->NumFloatLines++;

}

/* check */
if(Pfr->NumLines!=(Pfr->NumFloatLines+Pfr->NumAlphaLines))Quit(PARF_ERR,Pfr->ParameterFile);



/* Alloc memory for FloatLines and AlphaLines ID vectors */
if(AllocFlag){
Pfr->FloatLines=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->FloatLines)Quit(MEMORY_ERR,"");

Pfr->AlphaLines=(unsigned *)calloc(Pfr->NumAlphaLines,sizeof(unsigned));
if(!Pfr->AlphaLines)Quit(MEMORY_ERR,"");
}

/* Identify FloatLines and AlphaLines indexes from char Sheet */
j=k=0;
for(i=0;i<Pfr->NumLines;i++){

if(NumberOfFloatsInLine(Pfr->Sheet[i])==0)Pfr->AlphaLines[j++]=i;
else Pfr->FloatLines[k++]=i;

}

/* Alloc memory for float FloatDat row pointers */
if(AllocFlag){
Pfr->FloatDat=(float **)calloc(Pfr->NumFloatLines,sizeof(float *));
if(!Pfr->FloatDat)Quit(MEMORY_ERR,"");

/* Alloc memory for alphanumeric data */
Pfr->AlphaDat=CharMatCAlloc(Pfr->NumAlphaLines,MAXARGLEN);

/* Alloc memory for the array that contains the size of each row of FloatDat */
Pfr->VectSize=(unsigned *)calloc(Pfr->NumFloatLines,sizeof(unsigned));
if(!Pfr->VectSize)Quit(MEMORY_ERR,"");
}

/* obtain float FloatDat sheet from char Sheet */
for(i=0;i<Pfr->NumFloatLines;i++){

if(AllocFlag)Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],NULL);
else Pfr->FloatDat[i]=GetFloatVectFromLine(Pfr->Sheet[Pfr->FloatLines[i]],&Pfr->VectSize[i],Pfr->FloatDat[i]);

if(Pfr->FloatDat[i]==NULL)Quit(PARF_ERR,Pfr->ParameterFile);
}

/* obtain string AlphaDat sheet from first string in Sheet line */
for(i=0;i<Pfr->NumAlphaLines;i++){
sscanf(Pfr->Sheet[Pfr->AlphaLines[i]],"%s",Pfr->AlphaDat[i]);
}


}



/**************************************************************************
* FUNC: void NetRead(NEURALNET *Net,char *FName)
*
* DESC: Read the MLP specification from an ascii text file FName.
*************************************************************************/
void NetRead(NEURALNET *Net,char *FName)
{

#define INEQUALITY_DET_TRIGGER 0.01

register unsigned i,j,k,LineCount;

static PARFR PfrS;
static char EntryFlag;
char AllocFlag;
float *FTmpPtr;


unsigned NumLayers,NumInputNodes;


/*  AllocFlag=1 on the first call of NetRead(). AllocFlag=0 in
any other instant */
if(!EntryFlag){EntryFlag=AllocFlag=1;}
else {AllocFlag=0;}

/* read neural network from File */
Automatic_ParamFileReader(&PfrS,FName,AllocFlag);


/* Get architecture parameters */
LineCount=0;

NumLayers=PfrS.FloatDat[LineCount++][0]; /* LineCount = 0 */

NumInputNodes=PfrS.FloatDat[LineCount++][0]; /* LineCount = 1 */

if((NumLayers!=Net->NumLayers)||NumInputNodes!=Net->NumInputNodes)Quit(READ_NNET_ERR,FName);

Net->Phy_A=PfrS.FloatDat[LineCount][0]; /* LineCount = 2 */
Net->Phy_B=PfrS.FloatDat[LineCount++][1]; /* LineCount = 2 */

Net->NumEpoch=PfrS.FloatDat[LineCount++][0]; /* LineCount = 3 */

Net->MSE_T=PfrS.FloatDat[LineCount++][0]; /* LineCount = 4 */

Net->MSE_V=PfrS.FloatDat[LineCount++][0]; /* LineCount = 5 */

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 6 */

/* check architecture compatibility */
if(((unsigned)*(FTmpPtr++))!=Net->NumInputNodes)Quit(READ_NNET_ERR,FName);
for(i=0;i<Net->NumLayers;i++){
if(Net->Top[i]!=(unsigned)FTmpPtr[i])Quit(READ_NNET_ERR,FName);
}

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 7 */

for(i=0;i<Net->NumInputNodes;i++){
if(fabs(1-Net->Avg[i]/FTmpPtr[i])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
}

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 8 */

for(i=0;i<Net->NumInputNodes;i++){
if(fabs(1-Net->SDev[i]/FTmpPtr[i])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
}

FTmpPtr=PfrS.FloatDat[LineCount++];/* LineCount = 9 */

if(fabs(1-Net->A_n/FTmpPtr[0])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);
if(fabs(1-Net->B_n/FTmpPtr[1])>INEQUALITY_DET_TRIGGER)Quit(READ_NNET_ERR,FName);

/* i ->  Layer index */
/* j ->  Neuron index */
/* k ->  Synapse Weight index NOTE: W[0]=>bias; */


for(i=0;i<Net->NumLayers;i++){
for(j=0;j<Net->Top[i];j++){

FTmpPtr=PfrS.FloatDat[LineCount]; /* LineCount = 10 + j */

if(i==0){ /* input layer */


for(k=0;k<=Net->NumInputNodes;k++){  /* <= is due to bias synapse */
Net->L[i].N[j].W[k]=FTmpPtr[k];
Net->L[i].N[j].DW_1[k]=0; /* zeroes previous synaptic weights correction factors vector: */
}/* for k */

}/* if i==0 */


else{ /* NOT input layer */


for(k=0;k<=Net->Top[i-1];k++){  /* <= is due to bias synapse */
Net->L[i].N[j].W[k]=FTmpPtr[k]; /* LineCount = 10 + j */
Net->L[i].N[j].DW_1[k]=0; /* zeroes previous synaptic weights correction factors vector: */
}/* for k */

}/* else */

LineCount++; /* LineCount increment = j increment = neuron index increment */

}/* for j */


}/* for i */


}




/***************************************************************************
* FUNC: void NetWrite(NEURALNET *Net, char *FName)
*
* DESC: Write MLP specification in file FName.
****************************************************************************/
void NetWrite(NEURALNET *Net, char *FName)
{

char ParFileTimeDate[128];

FILE *fp;

register unsigned i; /*  Layer index */
register unsigned j; /*  Neuron index */
register unsigned k; /*  Synapse Weight index NOTE: W[0]=>bias; */


if((fp=fopen(FName,"wt"))==NULL)Quit(WRITE_OPEN_ERR,FName);
if (setvbuf(fp, Net->buffer, _IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,FName);

fprintf(fp,"%8s\t%c MLPTRV parameter file and its date & time.\n","ParFileTimeDate",'%');
fprintf(fp,"%8u\t%c The number of computational layers in the MLP.\n",Net->NumLayers,'%');
fprintf(fp,"%8u\t%c The number of input nodes in the MLP.\n",Net->NumInputNodes,'%');
fprintf(fp,"%8.6G\t%8.6G\t%c Activation function Phy(v)=A*tanh(B*v) parameters A and B.\n",Net->Phy_A,Net->Phy_B,'%');
fprintf(fp,"%8lu\t%c The final number of epochs.\n",Net->NumEpoch,'%');
fprintf(fp,"%8.6G\t%c The final training MSE.\n",Net->MSE_T,'%');
fprintf(fp,"%8.6G\t%c The final validation MSE.\n",Net->MSE_V,'%');
fprintf(fp,"The MLP architecture (leftmost - the number of input nodes / rightmost - the number of output neurons):\n");
fprintf(fp,"%u\t",Net->NumInputNodes);
for(i=0;i<Net->NumLayers;i++)fprintf(fp,"%u\t",Net->Top[i]);
fprintf(fp,"\nThe training set average vector:\n");
for(i=0;i<Net->NumInputNodes;i++)fprintf(fp,"%8.6G\t",Net->Avg[i]);
fprintf(fp,"\nThe training set standard deviation vector:\n");
for(i=0;i<Net->NumInputNodes;i++)fprintf(fp,"%8.6G\t",Net->SDev[i]);
fprintf(fp,"\nThe training desired output set D normalization parameters A_n and B_n, such that Dn=A_n*D+B_n where Dn lies in the range [-1,1]:\n");
fprintf(fp,"%8.6G\t%8.6G\t",Net->A_n,Net->B_n);
fprintf(fp,"\nEach following numeric line represents one MLP synapses vector (leftmost component -> bias synapse):\n");
for(i=0;i<Net->NumLayers;i++){
fprintf(fp,"Synapses vectors of each one of the %u neurons in computational layer %u:\n",Net->Top[i],i);
for(j=0;j<Net->Top[i];j++){

if(i==0){ /* if first Layer */

for(k=0;k<=Net->NumInputNodes;k++){  /* <= is due to bias synapse */
fprintf(fp,"%8.6G\t",Net->L[i].N[j].W[k]);
}/* for k */


}/* if i==0 */


else{  /* if not first Layer */


for(k=0;k<=Net->Top[i-1];k++){  /* <= is due to bias synapse */
fprintf(fp,"%8.6G\t",Net->L[i].N[j].W[k]);
}/* for k */


}/* else */

fprintf(fp,"\n");

}/* for j */

//fprintf(fp,"\n");

}/* for i */

fclose(fp);

}


/*****************************************************************************
* FUNC: void MLPTrainer(NEURALNET *Net)
*
* DESC: MLP training/validation loop.
*****************************************************************************/
void MLPTrainer(NEURALNET *Net)
{

#define RECTANGLE 22
#define DOWN_TRIANGLE 31
#define UP_TRIANGLE 30


register unsigned i;

float MSE_T_1=0; /* previous epoch Net->MSE_T */
float MSE_V_1=0; /* previous epoch Net->MSE_V */
float MSE_T_ChgRate; /* Training MSE changing rate */
float MSE_V_ChgRate; /* Validation MSE changing rate */
register unsigned long Epoch;  /* epoch counter */
unsigned SatDet; /* neuron saturation detector */
unsigned MSE_V_LockUpDet; /* validaton MSE lock up detector */
unsigned SmallEnoughMSEChgRateCounter; /* small enough MSE change rate counter  */
char MSEStr[128]; /* string to store the MSE value */

/* Open MSE output file */
if((Net->MSEOFp=fopen(Net->MSEOF,"wt"))==NULL)Quit(WRITE_OPEN_ERR,Net->MSEOF);
if (setvbuf(Net->MSEOFp,NULL,_IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Net->MSEOF);

/* Init variables */
Epoch=SatDet=MSE_V_LockUpDet=SmallEnoughMSEChgRateCounter=0;

/* Training/validation loop */
while(1){

/* Initial procedures */
gotoxy(XPos[0],YPos[0]);cprintf("%3lu",Epoch); /* show in Status screen */

Net->NumEpoch=Epoch;

if(Epoch==0){
MSE_V_1=Net->MSE_V;
MSE_T_1=Net->MSE_T;
}

/* Training Set presentation */
Net->MSE_T=0;
for(i=0;i<Net->TSSize;i++)
{
Net->MSE_T+=ForwardPass(Net, Net->TSet[i]);
BackwardPass(Net, Net->TSet[i]);
}

Net->MSE_T/=Net->TSSize;
MSE_T_ChgRate=(Net->MSE_T/MSE_T_1)-1;

/* Validation Set presentation */
if(Net->ValidationMode){
Net->MSE_V=0;
for(i=0;i<Net->VSSize;i++)
{
Net->MSE_V+=ForwardPass(Net, Net->VSet[i]);
}

Net->MSE_V/=Net->VSSize;
MSE_V_ChgRate=(Net->MSE_V/MSE_V_1)-1;
}
else{
Net->MSE_V=Net->MSE_T;
MSE_V_ChgRate=MSE_T_ChgRate;
}


/* Write Training & Validation MSE to Net->MSEOFp */
if(Net->ValidationMode)fprintf(Net->MSEOFp,"%-.1E\t%-.1E\n",Net->MSE_T,Net->MSE_V);
else fprintf(Net->MSEOFp,"%-.1E\n",Net->MSE_V);


/* Show in Status screen: */
gotoxy(XPos[2],YPos[2]);cprintf("%-.1E",Net->MSE_T);
gotoxy(XPos[3],YPos[3]);cprintf("%-.1E\r",Net->MSE_V);
gotoxy(XPos[4],YPos[4]);;cprintf(" %-8.2G%c\r",fabs(100*MSE_V_ChgRate),'%');


/* Show validation MSE motion indicator in Status screen */
gotoxy(XPos[1],YPos[1]);
if(MSE_V_ChgRate==0.0){
textcolor(YELLOW);
cprintf("%c",RECTANGLE);
}
else if(MSE_V_ChgRate<0){
textcolor(LIGHTGREEN);
cprintf("%c",DOWN_TRIANGLE);
}
else
{
textcolor(LIGHTRED);
cprintf("%c",UP_TRIANGLE);
}
textcolor(WHITE);


/* Update neurons saturation detector */
if(MSE_T_1==Net->MSE_T){SatDet++;}
else{
if(SatDet){SatDet--;}
}

/* Update validation MSE_lock up detector */
if(MSE_V_1==Net->MSE_V){MSE_V_LockUpDet++;}
else{
if(MSE_V_LockUpDet){MSE_V_LockUpDet--;}
}

/* Count the number of nearly consecutive epochs for which the validation MSE is
reducing (MSE_V_ChgRate<0) in a rate  smaller than Net->MSEChRStopCrit */
if((MSE_V_ChgRate<0)&&(fabs(100*MSE_V_ChgRate)<Net->MSEChRStopCrit)){SmallEnoughMSEChgRateCounter++;}
else{
if(SmallEnoughMSEChgRateCounter){SmallEnoughMSEChgRateCounter--;}
}


/* Terminate training as a consequence of all neurons saturation */
if(SatDet>Net->SmMSEChNEp2C){
QuitPreProcedure(Net,MSEStr);
Quit(SAT_INT,MSEStr);/* training terminated by neurons saturation */
}

/* Convergence is assumed if the number of nearly consecutive epochs for which the
validation MSE is reducing (MSE_V_ChgRate<0) in a rate  smaller than Net->MSEChRStopCrit
is greater than Net->SmMSEChNEp2C OR if validation MSE is locked up */
if((SmallEnoughMSEChgRateCounter>Net->SmMSEChNEp2C)||(MSE_V_LockUpDet>Net->SmMSEChNEp2C)){
QuitPreProcedure(Net,MSEStr);
Quit(SMALL_MSECHG_INT,MSEStr); /* training terminated by small validation MSE change rate */
}

/* Convergence is assumed if validation MSE if small enough */
if(Net->MSE_V<Net->MSEValStopCrit){
QuitPreProcedure(Net,MSEStr);
Quit(SMALL_MSE_INT,MSEStr); /* training terminated by small validation MSE */
}

/* Shuffle Training Set */
RowShuffle(Net->TSet,Net->TSSize);

/* If keyboard hit, call ChangeParameterMenu() menu */
if(kbhit()){
if(ChangeParameterMenu(Net)=='R'){Epoch=Net->NumEpoch;}
ShowParameters(Net);
NetWrite(Net,Net->FMLPSF);
}

/* Update loop variables */
MSE_T_1=Net->MSE_T;
MSE_V_1=Net->MSE_V;
Epoch++;

}/* while(1) */

}


/***************************************************************************
* FUNC: void TSetDeAvg (NEURALNET *Net)
*
* DESC: Subtract the average vector Avg[] from training set TSet.
*       Stores Avg[] in Net. The desired output columns of TSet are
*       left unchanged.
****************************************************************************/
void TSetDeAvg (NEURALNET *Net)
{

register unsigned i,j;



for(j=0;j<Net->NumInputNodes;j++){
Net->Avg[j]=0;

for(i=0;i<Net->TSSize;i++){
Net->Avg[j]+=Net->TSet[i][j];
}

Net->Avg[j]/=Net->TSSize;  /* the average value of the j_th TSet column */

for(i=0;i<Net->TSSize;i++){
Net->TSet[i][j]-=Net->Avg[j]; /* de-averages TSet */
}


}


}


/***************************************************************************
* FUNC: void TSetDeSDev(NEURALNET *Net)
*
* DESC: Normalize each column of training set TSet by its standard deviation
*       SDev. Stores SDev[] in Net. The desired output columns of TSet are
*       left unchanged.
****************************************************************************/
void TSetDeSDev(NEURALNET *Net)
{

register unsigned i,j;



for(j=0;j<Net->NumInputNodes;j++){
Net->SDev[j]=0;

for(i=0;i<Net->TSSize;i++){
Net->SDev[j]+=SQR(Net->TSet[i][j]);
}

Net->SDev[j]/=Net->TSSize;

Net->SDev[j]=sqrt(Net->SDev[j]); /* the standard deviation of the j_th TSet column */

if(Net->SDev[j]){
for(i=0;i<Net->TSSize;i++){
Net->TSet[i][j]/=Net->SDev[j]; /* normalizes TSet by SDev */
}
}


}


}



/*****************************************************************************
* FUNC: void NormalizeDesiredOutput(NEURALNET *Net)
*
* DESC: Determine the desired output set D normalization parameters A_n and B_n,
*       such that Dn=A_n*D+B_n where Dn lies in the range [-1,1]. Then, normalize
*       D to the range [-1,1], i.e., D=Dn.
*****************************************************************************/
void NormalizeDesiredOutput(NEURALNET *Net)
{

register unsigned i,j;
float v;

float maxv,minv;

maxv=-MAXFLOAT;
minv=MAXFLOAT;

/* determine maxv and minv of desired output set D */
for(i=0;i<Net->TSSize;i++){
for(j=Net->NumInputNodes;j<Net->NumInputNodes+Net->Top[(Net->NumLayers)-1];j++){

v=Net->TSet[i][j];
if(v>maxv)maxv=v;
if(v<minv)minv=v;

}
}

/* determine normalization parameters */
Net->A_n=2.0/(maxv-minv);
Net->B_n=(maxv+minv)/(minv-maxv);

/* normalize desired output set */
for(i=0;i<Net->TSSize;i++){
for(j=Net->NumInputNodes;j<Net->NumInputNodes+Net->Top[(Net->NumLayers)-1];j++){

Net->TSet[i][j]=Net->A_n*Net->TSet[i][j]+Net->B_n;

}
}



}


/***************************************************************************
* FUNC: void VSetDeAvg (NEURALNET *Net)
*
* DESC: Subtract the average vector Avg[] from training validation VSet.
*       The desired output columns of VSet are left unchanged.
****************************************************************************/
void VSetDeAvg (NEURALNET *Net)
{

register unsigned i,j;


for(j=0;j<Net->NumInputNodes;j++){

for(i=0;i<Net->VSSize;i++){
Net->VSet[i][j]-=Net->Avg[j]; /* de-averages VSet */
}


}


}


/***************************************************************************
* FUNC: void VSetDeSDev(NEURALNET *Net)
*
* DESC: Normalize each column of validation set VSet by the corresponding
*       column standard deviation SDev in TSet. The desired output columns
*       of VSet are left unchanged.
****************************************************************************/
void VSetDeSDev(NEURALNET *Net)
{

register unsigned i,j;


for(j=0;j<Net->NumInputNodes;j++){


if(Net->SDev[j]){
for(i=0;i<Net->VSSize;i++){
Net->VSet[i][j]/=Net->SDev[j]; /* normalizes VSet by TSet SDev */
}
}


}


}


/*****************************************************************************
* FUNC: void NormalizeVSetDesiredOutput(NEURALNET *Net)
*
* DESC: Normalize VSet desired output set to the range [-1,1].
*****************************************************************************/
void NormalizeDesiredVSetOutput(NEURALNET *Net)
{

register unsigned i,j;

/* normalize VSet desired output set */
for(i=0;i<Net->VSSize;i++){
for(j=Net->NumInputNodes;j<Net->NumInputNodes+Net->Top[(Net->NumLayers)-1];j++){

Net->VSet[i][j]=Net->A_n*Net->VSet[i][j]+Net->B_n;

}
}



}


/*****************************************************************************
* FUNC: void QuitPreProcedure(NEURALNET *Net, char *MSE2String)
*
* DESC: Perform pre-exit procedure before the call of Quit() due to training/validation
*       termination.
*****************************************************************************/
void QuitPreProcedure(NEURALNET *Net, char *MSE2String)

{

register unsigned i,j;
unsigned nrows,ncols_;

/* close MSE output file */
fclose(Net->MSEOFp);

/* save specification after training */
NetWrite(Net,Net->FMLPSF);

/* save Net->MSE_V into MSE2String */
sprintf(MSE2String," %-.3E",Net->MSE_V);

if(!Net->ValidationMode){ /* -> Net->VSet==Net->TSet */

/* Re-Read Training Set (to circumvent RowShuffle() action) */
Net->TSet=FloatMatReadCheck(Net->TSet,Net->TDIF,Net->buffer,&nrows,&ncols_,0);

/*  Normalize TSet(=VSet) desired output set to range [-1,1] */
NormalizeDesiredVSetOutput(Net);

/* De-averages Training Set */
VSetDeAvg(Net);

/* Normalizes each TSet input vector component by the corresponding
component of the TSet standard deviation vector. */
VSetDeSDev(Net);

}

/* Open MLP output neurons output file */
if((Net->MLPO2VOFp=fopen(Net->MLPO2VOF,"wt"))==NULL)Quit(WRITE_OPEN_ERR,Net->MLPO2VOF);
if (setvbuf(Net->MLPO2VOFp,Net->buffer,_IOFBF, BUFSIZE) != 0)Quit(BUFFER_ERR,Net->MLPO2VOF);


/* Validation Set presentation */
for(i=0;i<Net->VSSize;i++){

ForwardPass(Net, Net->VSet[i]);

/* Save Output Neurons values */
for(j=0;j<(Net->Top[(Net->NumLayers)-1]);j++)fprintf(Net->MLPO2VOFp,"%8.6G\t",(Net->L[(Net->NumLayers)-1].N[j].v-Net->B_n)/Net->A_n);
fprintf(Net->MLPO2VOFp,"\n");

}

fclose(Net->MLPO2VOFp);


//Beep(1000,200);


}
